import time
from copy import deepcopy
from typing import Dict, List, Optional

import torch
import numpy as np
import pandas as pd
from torch import nn
import torch_optimizer as ad_optim
from sklearn.metrics import r2_score
from torch.cuda.amp import GradScaler
from torch.utils.data import DataLoader
from torch.nn.modules.loss import MSELoss
from torch.optim.optimizer import Optimizer

from type_infer.dtype import dtype
from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.data.encoded_ds import EncodedDs
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.helpers.ar_net import ArNet
from lightwood.mixer.helpers.default_net import DefaultNet
from lightwood.api.types import PredictionArguments
from lightwood.mixer.helpers.transform_corss_entropy_loss import TransformCrossEntropyLoss


class Neural(BaseMixer):
    model: nn.Module
    dtype_dict: dict
    target: str
    epochs_to_best: int
    fit_on_dev: bool
    supports_proba: bool

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            target_encoder: BaseEncoder,
            net: str,
            fit_on_dev: bool,
            search_hyperparameters: bool,
            n_epochs: Optional[int] = None
    ):
        """
        The Neural mixer trains a fully connected dense network from concatenated encoded outputs of each of the features in the dataset to predicted the encoded output. 
        
        :param stop_after: How long the total fitting process should take
        :param target: Name of the target column
        :param dtype_dict: Data type dictionary
        :param target_encoder: Reference to the encoder used for the target
        :param net: The network type to use (`DeafultNet` or `ArNet`)
        :param fit_on_dev: If we should fit on the dev dataset
        :param search_hyperparameters: If the network should run a more through hyperparameter search (currently disabled)
        :param n_epochs: amount of epochs that the network will be trained for. Supersedes all other early stopping criteria if specified.
        """ # noqa
        super().__init__(stop_after)
        self.dtype_dict = dtype_dict
        self.target = target
        self.target_encoder = target_encoder
        self.epochs_to_best = 0
        self.n_epochs = n_epochs
        self.fit_on_dev = fit_on_dev
        self.net_class = DefaultNet if net == 'DefaultNet' else ArNet
        self.supports_proba = dtype_dict[target] in [dtype.binary, dtype.categorical]
        self.search_hyperparameters = search_hyperparameters
        self.stable = True

    def _final_tuning(self, data):
        if self.dtype_dict[self.target] in (dtype.integer, dtype.float, dtype.quantity):
            self.model = self.model.eval()
            with torch.no_grad():
                acc_dict = {}
                for decode_log in [True, False]:
                    self.target_encoder.decode_log = decode_log
                    decoded_predictions = []
                    decoded_real_values = []
                    for X, Y in data:
                        X = X.to(self.model.device)
                        Y = Y.to(self.model.device)
                        Yh = self.model(X)

                        Yh = torch.unsqueeze(Yh, 0) if len(Yh.shape) < 2 else Yh
                        Y = torch.unsqueeze(Y, 0) if len(Y.shape) < 2 else Y

                        decoded_predictions.extend(self.target_encoder.decode(Yh))
                        decoded_real_values.extend(self.target_encoder.decode(Y))

                    acc_dict[decode_log] = r2_score(decoded_real_values, decoded_predictions)

            self.target_encoder.decode_log = acc_dict[True] > acc_dict[False]

    def _select_criterion(self) -> torch.nn.Module:
        if self.dtype_dict[self.target] in (dtype.categorical, dtype.binary):
            criterion = TransformCrossEntropyLoss(weight=self.target_encoder.index_weights.to(self.model.device))
        elif self.dtype_dict[self.target] in (dtype.tags, dtype.cat_tsarray):
            criterion = nn.BCEWithLogitsLoss()
        elif self.dtype_dict[self.target] in (dtype.cat_array, ):
            criterion = nn.L1Loss()
        elif self.dtype_dict[self.target] in (dtype.integer, dtype.float, dtype.quantity, dtype.num_array):
            criterion = MSELoss()
        else:
            criterion = MSELoss()

        return criterion

    def _select_optimizer(self) -> Optimizer:
        optimizer = ad_optim.Ranger(self.model.parameters(), lr=self.lr, weight_decay=2e-2)
        return optimizer

    def _find_lr(self, dl):
        optimizer = self._select_optimizer()
        criterion = self._select_criterion()
        scaler = GradScaler()

        running_losses: List[float] = []
        cum_loss = 0
        lr_log = []
        best_model = self.model
        stop = False
        batches = 0
        for epoch in range(1, 101):
            if stop:
                break

            for i, (X, Y) in enumerate(dl):
                if stop:
                    break

                batches += len(X)
                X = X.to(self.model.device)
                Y = Y.to(self.model.device)
                with LightwoodAutocast():
                    optimizer.zero_grad()
                    Yh = self.model(X)
                    loss = criterion(Yh, Y)
                    if LightwoodAutocast.active:
                        scaler.scale(loss).backward()
                        scaler.step(optimizer)
                        scaler.update()
                    else:
                        loss.backward()
                        optimizer.step()
                cum_loss += loss.item()

                # Account for ranger lookahead update
                if (i + 1) * epoch % 6:
                    batches = 0
                    lr = optimizer.param_groups[0]['lr']
                    log.info(f'Loss of {cum_loss} with learning rate {lr}')
                    running_losses.append(cum_loss)
                    lr_log.append(lr)
                    cum_loss = 0
                    if len(running_losses) < 2 or np.mean(running_losses[:-1]) > np.mean(running_losses):
                        optimizer.param_groups[0]['lr'] = lr * 1.4
                        # Time saving since we don't have to start training fresh
                        best_model = deepcopy(self.model)
                    else:
                        stop = True

        best_loss_lr = lr_log[np.argmin(running_losses)]
        lr = best_loss_lr
        log.info(f'Found learning rate of: {lr}')
        return lr, best_model

    def _max_fit(self, train_dl, dev_dl, criterion, optimizer, scaler, stop_after, return_model_after):
        epochs_to_best = 0
        best_dev_error = pow(2, 32)
        running_errors = []
        best_model = self.model

        for epoch in range(1, return_model_after + 1):
            self.model = self.model.train()
            running_losses: List[float] = []
            for i, (X, Y) in enumerate(train_dl):
                X = X.to(self.model.device)
                Y = Y.to(self.model.device)
                with LightwoodAutocast():
                    optimizer.zero_grad()
                    Yh = self.model(X)
                    loss = criterion(Yh, Y)
                    if LightwoodAutocast.active:
                        scaler.scale(loss).backward()
                        scaler.step(optimizer)
                        scaler.update()
                    else:
                        loss.backward()
                        optimizer.step()

                running_losses.append(loss.item())
                if (time.time() - self.started) > stop_after:
                    break

            train_error = np.mean(running_losses)
            epoch_error = self._error(dev_dl, criterion)
            running_errors.append(epoch_error)
            log.info(f'Loss @ epoch {epoch}: {epoch_error}')

            if np.isnan(train_error) or np.isnan(
                    running_errors[-1]) or np.isinf(train_error) or np.isinf(
                    running_errors[-1]):
                break

            if best_dev_error > running_errors[-1]:
                best_dev_error = running_errors[-1]
                best_model = deepcopy(self.model)
                epochs_to_best = epoch

            # manually set epoch limit
            if self.n_epochs is not None:
                if epoch > self.n_epochs:
                    break

            # automated early stopping
            else:
                if len(running_errors) >= 5:
                    delta_mean = np.average([running_errors[-i - 1] - running_errors[-i] for i in range(1, 5)],
                                            weights=[(1 / 2)**i for i in range(1, 5)])
                    if delta_mean <= 0:
                        break
                elif (time.time() - self.started) > stop_after:
                    break
                elif running_errors[-1] < 0.0001 or train_error < 0.0001:
                    break

        if np.isnan(best_dev_error):
            best_dev_error = pow(2, 32)
        return best_model, epochs_to_best, best_dev_error

    def _error(self, dev_dl, criterion) -> float:
        self.model = self.model.eval()
        running_losses: List[float] = []
        with torch.no_grad():
            for X, Y in dev_dl:
                X = X.to(self.model.device)
                Y = Y.to(self.model.device)
                Yh = self.model(X)
                running_losses.append(criterion(Yh, Y).item())
            return np.mean(running_losses)

    def _init_net(self, ds: EncodedDs):
        net_kwargs = {'input_size': len(ds[0][0]),
                      'output_size': len(ds[0][1]),
                      'num_hidden': self.num_hidden,
                      'dropout': 0}

        if self.net_class == ArNet:
            net_kwargs['encoder_span'] = ds.encoder_spans
            net_kwargs['target_name'] = self.target

        self.model = self.net_class(**net_kwargs)

    # @TODO: Compare partial fitting fully on and fully off on the benchmarks!
    # @TODO: Writeup on the methodology for partial fitting
    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the Neural mixer on some data, making it ready to predict

        :param train_data: The network is fit/trained on this
        :param dev_data: Data used for early stopping and hyperparameter determination
        """
        # ConcatedEncodedDs
        self.started = time.time()
        self.batch_size = min(200, int(len(train_data) / 10))
        self.batch_size = max(40, self.batch_size)

        dev_dl = DataLoader(dev_data, batch_size=self.batch_size, shuffle=False)
        train_dl = DataLoader(train_data, batch_size=self.batch_size, shuffle=False)

        self.lr = 1e-4
        self.num_hidden = 1

        # Find learning rate
        # keep the weights
        self._init_net(train_data)
        self.lr, self.model = self._find_lr(train_dl)

        # Keep on training
        optimizer = self._select_optimizer()
        criterion = self._select_criterion()
        scaler = GradScaler()

        # Only 0.8 of the remaining time budget is used to allow some time for the final tuning and partial fit
        self.model, epoch_to_best_model, _ = self._max_fit(
            train_dl, dev_dl, criterion, optimizer, scaler, (self.stop_after - (time.time() - self.started)) * 0.8,
            return_model_after=20000)

        self.epochs_to_best += epoch_to_best_model

        if self.fit_on_dev:
            self.partial_fit(dev_data, train_data)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        self._fit(train_data, dev_data)
        self._final_tuning(dev_data)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        """
        Augments the mixer's fit with new data, nr of epochs is based on the amount of epochs the original fitting took

        :param train_data: The network is fit/trained on this
        :param dev_data: Data used for early stopping and hyperparameter determination
        """

        # Based this on how long the initial training loop took, at a low learning rate as to not mock anything up tooo badly # noqa
        self.started = time.time()
        train_dl = DataLoader(train_data, batch_size=self.batch_size, shuffle=True)
        dev_dl = DataLoader(dev_data, batch_size=self.batch_size, shuffle=True)
        optimizer = self._select_optimizer()
        criterion = self._select_criterion()
        scaler = GradScaler()

        self.model, _, _ = self._max_fit(train_dl, dev_dl, criterion, optimizer, scaler,
                                         self.stop_after * 0.1, max(1, int(self.epochs_to_best / 3)))

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Make predictions based on datasource with the same features as the ones used for fitting

        :param ds: Predictions are generate from it
        :param arg: Any additional arguments used in predicting

        :returns: A dataframe cotaining the decoded predictions and (depending on the args) additional information such as the probabilites for each target class
        """ # noqa
        self.model = self.model.eval()
        decoded_predictions: List[object] = []
        all_probs: List[List[float]] = []
        rev_map = {}

        with torch.no_grad():
            for idx, (X, Y) in enumerate(ds):
                X = X.to(self.model.device)
                Yh = self.model(X)
                Yh = torch.unsqueeze(Yh, 0) if len(Yh.shape) < 2 else Yh

                kwargs = {}
                for dep in self.target_encoder.dependencies:
                    kwargs['dependency_data'] = {dep: ds.data_frame.iloc[idx][[dep]].values}

                if args.predict_proba and self.supports_proba:
                    decoded_prediction, probs, rev_map = self.target_encoder.decode_probabilities(Yh, **kwargs)
                    all_probs.append(probs)
                else:
                    decoded_prediction = self.target_encoder.decode(Yh, **kwargs)

                decoded_predictions.extend(decoded_prediction)

            ydf = pd.DataFrame({'prediction': decoded_predictions})

            if args.predict_proba and self.supports_proba:
                raw_predictions = np.array(all_probs).squeeze(axis=1)

                for idx, label in enumerate(rev_map.values()):
                    ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

            return ydf



from lightwood.helpers.accuracy import to_binary, f1_score, recall_score, precision_score, r2_score
from lightwood.helpers.device import is_cuda_compatible, get_devices
from lightwood.helpers.general import mase, is_none, evaluate_accuracy, evaluate_num_array_accuracy,\
    evaluate_array_accuracy, evaluate_multilabel_accuracy, evaluate_regression_accuracy, evaluate_cat_array_accuracy, \
    bounded_ts_accuracy
from lightwood.helpers.ts import get_group_matches, get_ts_groups, get_inferred_timestamps, add_tn_num_conf_bounds, \
    add_tn_cat_conf_bounds
from lightwood.helpers.io import read_from_path_or_url
from lightwood.helpers.parallelism import get_nr_procs, mut_method_call, run_mut_method
from lightwood.helpers.numeric import filter_nan_and_none
from lightwood.helpers.seed import seed
from lightwood.helpers.text import get_pct_auto_increment, extract_digits, isascii,\
    hashtext, splitRecursive, gen_chars, word_tokenize
from lightwood.helpers.torch import average_vectors, concat_vectors_and_pad, LightwoodAutocast


__all__ = ['to_binary', 'f1_score', 'recall_score', 'precision_score', 'r2_score', 'is_cuda_compatible', 'get_devices',
           'get_group_matches', 'get_ts_groups', 'mase', 'is_none', 'evaluate_accuracy', 'evaluate_num_array_accuracy',
           'evaluate_array_accuracy', 'evaluate_cat_array_accuracy', 'bounded_ts_accuracy', 'get_pct_auto_increment',
           'evaluate_multilabel_accuracy', 'evaluate_regression_accuracy', 'read_from_path_or_url', 'get_nr_procs',
           'mut_method_call', 'run_mut_method',
           'extract_digits', 'isascii', 'get_inferred_timestamps', 'add_tn_num_conf_bounds', 'add_tn_cat_conf_bounds',
           'hashtext', 'splitRecursive', 'gen_chars', 'word_tokenize',
           'average_vectors', 'concat_vectors_and_pad', 'LightwoodAutocast', 'filter_nan_and_none', 'seed']  # noqa



import math
from typing import Iterable, List, Union
import torch
import numpy as np
from torch.types import Number
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log
from lightwood.helpers.general import is_none
from type_infer.dtype import dtype


class NumericEncoder(BaseEncoder):
    """
    The numeric encoder takes numbers (float or integer) and converts it into tensors of the form:
    ``[0 if the number is none, otherwise 1, 1 if the number is positive, otherwise 0, natural_log(abs(number)), number/absolute_mean]``

    This representation is: ``[1 if the number is positive, otherwise 0, natural_log(abs(number)), number/absolute_mean]]`` if encoding target values, since target values can't be none.

    The ``absolute_mean`` is computed in the ``prepare`` method and is just the mean of the absolute values of all numbers feed to prepare (which are not none)

    ``none`` stands for any number that is an actual python ``None`` value or any sort of non-numeric value (a string, nan, inf)
    """ # noqa

    def __init__(self, data_type: dtype = None, is_target: bool = False, positive_domain: bool = False):
        """
        :param data_type: The data type of the number (integer, float, quantity)
        :param is_target: Indicates whether the encoder refers to a target column or feature column (True==target)
        :param positive_domain: Forces the encoder to always output positive values
        """
        super().__init__(is_target)
        self._type = data_type
        self._abs_mean = None
        self.positive_domain = positive_domain
        self.decode_log = False
        self.output_size = 4 if not self.is_target else 3

    def prepare(self, priming_data: Iterable):
        """
        "NumericalEncoder" uses a rule-based form to prepare results on training (priming) data. The averages etc. are taken from this distribution.

        :param priming_data: an iterable data structure containing numbers numbers which will be used to compute the values used for normalizing the encoded representations
        """ # noqa
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        value_type = 'int'
        for number in priming_data:
            if not is_none(number):
                if int(number) != number:
                    value_type = 'float'

        self._type = value_type if self._type is None else self._type
        non_null_priming_data = [x for x in priming_data if not is_none(x)]
        self._abs_mean = np.mean(np.abs(non_null_priming_data))
        self.is_prepared = True

    def encode(self, data: Iterable):
        """
        :param data: An iterable data structure containing the numbers to be encoded

        :returns: A torch tensor with the representations of each number
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        ret = []
        for real in data:
            try:
                real = float(real)
            except Exception:
                real = None
            if self.is_target:
                # Will crash if ``real`` is not a float, this is fine, targets should always have a value
                vector = [0] * 3
                vector[0] = 1 if real < 0 and not self.positive_domain else 0
                vector[1] = math.log(abs(real)) if abs(real) > 0 else -20
                vector[2] = real / self._abs_mean

            else:
                vector = [0] * 4
                try:
                    if is_none(real):
                        vector[0] = 0
                    else:
                        vector[0] = 1
                        vector[1] = math.log(abs(real)) if abs(real) > 0 else -20
                        vector[2] = 1 if real < 0 and not self.positive_domain else 0
                        vector[3] = real / self._abs_mean
                except Exception as e:
                    vector = [0] * 4
                    log.error(f'Can\'t encode input value: {real}, exception: {e}')

            ret.append(vector)

        return torch.Tensor(ret)

    def decode(self, encoded_values: Union[List[Number], torch.Tensor], decode_log: bool = None) -> list:
        """
        :param encoded_values: The encoded values to decode into single numbers
        :param decode_log: Whether to decode the ``log`` or ``linear`` part of the representation, since the encoded vector contains both a log and a linear part

        :returns: The decoded number
        """ # noqa
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if decode_log is None:
            decode_log = self.decode_log

        ret = []
        if isinstance(encoded_values, torch.Tensor):
            encoded_values = encoded_values.tolist()

        for vector in encoded_values:
            if self.is_target:
                if np.isnan(
                        vector[0]) or vector[0] == float('inf') or np.isnan(
                        vector[1]) or vector[1] == float('inf') or np.isnan(
                        vector[2]) or vector[2] == float('inf'):
                    log.error(f'Got weird target value to decode: {vector}')
                    real_value = pow(10, 63)
                else:
                    if decode_log:
                        sign = -1 if vector[0] > 0.5 else 1
                        try:
                            real_value = math.exp(vector[1]) * sign
                        except OverflowError:
                            real_value = pow(10, 63) * sign
                    else:
                        real_value = vector[2] * self._abs_mean

                    if self.positive_domain:
                        real_value = abs(real_value)

                    if self._type == 'int':
                        real_value = int(real_value)

            else:
                if vector[0] < 0.5:
                    ret.append(None)
                    continue

                real_value = vector[3] * self._abs_mean

                if self._type == 'int':
                    real_value = round(real_value)

            if isinstance(real_value, torch.Tensor):
                real_value = real_value.item()
            ret.append(real_value)
        return ret



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import json
import hashlib
import nltk


try:
    nltk.data.find('tokenizers/punkt')
except LookupError:
    nltk.download('punkt')

try:
    from nltk.corpus import stopwords
    stopwords.words('english')
except LookupError:
    nltk.download('stopwords', quiet=True)


def word_tokenize(string):
    sep_tag = '{#SEP#}'
    for separator in WORD_SEPARATORS:
        string = str(string).replace(separator, sep_tag)

    words_split = string.split(sep_tag)
    num_words = len([word for word in words_split if word and word not in ['', None]])
    return num_words


def gen_chars(length, character):
    """
    # lambda to Generates a string consisting of `length` consiting of repeating `character`
    :param length:
    :param character:
    :return:
    """
    return ''.join([character for _ in range(length)])


def splitRecursive(word, tokens):
    words = [str(word)]
    for token in tokens:
        new_split = []
        for word in words:
            new_split += word.split(token)
        words = new_split
    words = [word for word in words if word not in ['', None]]
    return words


def hashtext(cell):
    text = json.dumps(cell)
    return hashlib.md5(text.encode('utf8')).hexdigest()


def isascii(string):
    """
    Used instead of str.isascii because python 3.6 doesn't have that
    """
    return all(ord(c) < 128 for c in string)


def extract_digits(point):
    return ''.join([char for char in str(point) if char.isdigit()])


def get_pct_auto_increment(data):
    int_data = []
    for point in [extract_digits(x) for x in data]:
        try:
            int_data.append(int(point))
        except Exception:
            pass

    int_data = sorted(int_data)
    prev_nr = int_data[0]
    increase_by_one = 0
    for nr in int_data[1:]:
        diff = nr - prev_nr
        if diff == 1:
            increase_by_one += 1
        prev_nr = nr

    return increase_by_one / (len(data) - 1)



"""
Time series utility methods for usage within mixers.
"""
from typing import Dict
from copy import deepcopy

import pandas as pd

from lightwood.data.encoded_ds import EncodedDs
from lightwood.api.types import TimeseriesSettings
from lightwood.helpers.ts import get_group_matches, get_ts_groups


def _transform_target(ts_analysis: Dict[str, Dict], df: pd.DataFrame, freq, group: tuple):
    df = deepcopy(df)  # needed because methods like get_level_values actually have undisclosed side effects  # noqa
    transformer = ts_analysis['stl_transforms'][group]['transformer']
    if isinstance(df.index, pd.MultiIndex) and len(df.index.levels) > 1:
        temp_s = df.droplevel(0)
        temp_s.index = pd.date_range(start=temp_s.index[0], freq=freq, periods=len(temp_s))
        return transformer.transform(temp_s.to_period())
    elif isinstance(df.index, pd.MultiIndex):
        temp_s = pd.Series(data=df.values)
        temp_s.index = df.index.levels[0]  # warning: do not use get_level_values as it removes inferred freq
        return transformer.transform(temp_s.to_period())
    else:
        return transformer.transform(df.to_period())


def _inverse_transform_target(ts_analysis: Dict[str, Dict], predictions: pd.DataFrame, freq, group: tuple):
    predictions = deepcopy(predictions)  # needed because methods like get_level_values actually have undisclosed side effects  # noqa
    transformer = ts_analysis['stl_transforms'][group]['transformer']
    if isinstance(predictions.index, pd.MultiIndex) and len(predictions.index.levels) > 1:
        temp_s = predictions.droplevel(0)
        temp_s.index = pd.date_range(start=temp_s.index[0], freq=freq, periods=len(temp_s))
        return transformer.inverse_transform(temp_s.to_period())
    elif isinstance(predictions.index, pd.MultiIndex):
        temp_s = pd.Series(data=predictions)
        temp_s.index = predictions.index.levels[0]  # warning: do not use get_level_values as it removes inferred freq
        return transformer.inverse_transform(temp_s.to_period())
    else:
        return transformer.inverse_transform(predictions.to_period())


def _apply_stl_on_training(
        train_data: EncodedDs,
        dev_data: EncodedDs,
        target: str,
        tss: TimeseriesSettings,
        ts_analysis: Dict[str, Dict],
):
    """
    The purpose of this function is a side-effect:
        applying STL blocks in the target column of the dataframes embedded
        in the passed EncodedDs instances.
        
    :param train_data: EncodedDs object for the training dataset 
    :param dev_data: EncodedDs object for the dev/validation dataset
    :return: 
    """  # noqa
    gby = tss.group_by if tss.group_by else []
    tr_midx = pd.MultiIndex.from_frame(train_data.data_frame.reset_index()[[*gby, 'index']])
    dev_midx = pd.MultiIndex.from_frame(dev_data.data_frame.reset_index()[[*gby, 'index']])
    train_data.data_frame.index = tr_midx
    dev_data.data_frame.index = dev_midx

    for group in ts_analysis['group_combinations']:
        if len(ts_analysis['group_combinations']) == 1 or group != '__default':
            freq = ts_analysis['sample_freqs'][group]
            train_idxs, train_subset = get_group_matches(train_data.data_frame, group, gby)
            dev_idxs, dev_subset = get_group_matches(dev_data.data_frame, group, gby)

            if train_subset.shape[0] > 0:
                train_data.data_frame[target].loc[train_idxs] = _transform_target(ts_analysis, train_subset[target],
                                                                                  freq, group).values
                for timestep in range(1, tss.horizon):
                    train_data.data_frame[f'{target}_timestep_{timestep}'].loc[train_idxs] = \
                        train_data.data_frame[target].loc[train_idxs].shift(-timestep)

            if dev_subset.shape[0] > 0:
                dev_data.data_frame[target].loc[dev_idxs] = _transform_target(ts_analysis, dev_subset[target],
                                                                              freq, group).values
                for timestep in range(1, tss.horizon):
                    dev_data.data_frame[f'{target}_timestep_{timestep}'].loc[dev_idxs] = \
                        dev_data.data_frame[target].loc[dev_idxs].shift(-timestep)


def _stl_transform(
        ydf: pd.DataFrame,
        ds: EncodedDs,
        target: str,
        tss: TimeseriesSettings,
        ts_analysis: Dict[str, Dict]
) -> pd.DataFrame:
    """
   The purpose of this function is a side-effect:
       applying STL blocks in the target column of the dataframes embedded
       in the passed EncodedDs instances.

    :param ydf: 
    :param ds: 
    :param target: 
    :param tss: 
    :param ts_analysis: 
    :return: 
    """  # noqa
    gby = tss.group_by if tss.group_by else []
    midx = pd.MultiIndex.from_frame(ds.data_frame.reset_index()[[*gby, 'index']])
    midx.levels[0].freq = ts_analysis['sample_freqs']['__default']  # note: assumes same freq across all groups
    ds.data_frame.index = midx
    ydf.index = midx
    groups = get_ts_groups(ds.data_frame, tss)
    for group in groups:
        group = group if group in ts_analysis['group_combinations'] else '__default'
        if len(ts_analysis['group_combinations']) == 1 or group != '__default':
            freq = ts_analysis['sample_freqs'][group]
            idxs, subset = get_group_matches(ds.data_frame, group, gby)
            ds.data_frame[target].loc[idxs] = _transform_target(ts_analysis, subset[target], freq, group).values
    return ds.data_frame  # TODO: check that the side effects actually worked outside the fn scope and avoid returning?


def _stl_inverse_transform(
        ydf: pd.DataFrame,
        ds: EncodedDs,
        tss: TimeseriesSettings,
        ts_analysis: Dict[str, Dict]
) -> None:
    """
    :param ydf: 
    :param ds: 
    :param target: 
    :param tss: 
    :param ts_analysis: 
    :return: 
    """  # noqa
    groups = get_ts_groups(ds.data_frame, tss)
    gby = tss.group_by if tss.group_by else []
    for group in groups:
        group = group if group in ts_analysis['group_combinations'] else '__default'
        if len(ts_analysis['group_combinations']) == 1 or group != '__default':
            idxs, subset = get_group_matches(ds.data_frame, group, gby)
            freq = ts_analysis['sample_freqs'][group]
            for timestep in range(tss.horizon):
                ydf[f'prediction_{timestep}'].loc[idxs] = _inverse_transform_target(ts_analysis,
                                                                                    ydf[f'prediction_{timestep}'].loc[
                                                                                        idxs],
                                                                                    freq,
                                                                                    group
                                                                                    ).values
    return ydf.reset_index(drop=True)



from copy import deepcopy
from typing import Dict, List, Union, Optional

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.mixer.helpers.ts import _apply_stl_on_training, _stl_transform, _stl_inverse_transform
from lightwood.encoder.base import BaseEncoder
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.lightgbm import LightGBM
from lightwood.api.types import PredictionArguments, TimeseriesSettings
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class LightGBMArray(BaseMixer):
    """LightGBM-based model, intended for usage in time series tasks."""
    models: List[LightGBM]
    submodel_stop_after: float
    target: str
    supports_proba: bool
    ts_analysis: Dict
    tss: TimeseriesSettings

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            input_cols: List[str],
            fit_on_dev: bool,
            target_encoder: BaseEncoder,
            ts_analysis: Dict[str, object],
            use_stl: bool,
            tss: TimeseriesSettings
    ):
        super().__init__(stop_after)
        self.tss = tss
        self.horizon = tss.horizon
        self.submodel_stop_after = stop_after / self.horizon
        self.target = target
        self.offset_pred_cols = [f'{self.target}_timestep_{i}' for i in range(1, self.horizon)]
        if set(input_cols) != {self.tss.order_by}:
            input_cols.remove(self.tss.order_by)
        for col in self.offset_pred_cols:
            dtype_dict[col] = dtype_dict[self.target]
        self.models = [LightGBM(self.submodel_stop_after,
                                target_col,
                                dtype_dict,
                                input_cols,
                                False,  # fit_on_dev,
                                True if tss.horizon < 10 else False,  # use_optuna
                                target_encoder)
                       for _, target_col in zip(range(self.horizon), [target] + self.offset_pred_cols)]
        self.ts_analysis = ts_analysis
        self.supports_proba = False
        self.use_stl = False
        self.stable = True

    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs, submodel_method='fit') -> None:
        original_train = deepcopy(train_data.data_frame)
        original_dev = deepcopy(dev_data.data_frame)

        if self.use_stl and self.ts_analysis.get('stl_transforms', False):
            _apply_stl_on_training(train_data, dev_data, self.target, self.tss, self.ts_analysis)

        for timestep in range(self.horizon):
            getattr(self.models[timestep], submodel_method)(train_data, dev_data)

        # restore dfs
        train_data.data_frame = original_train
        dev_data.data_frame = original_dev

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info('Started fitting LGBM models for array prediction')
        self._fit(train_data, dev_data, submodel_method='fit')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        log.info('Updating array of LGBM models...')
        self._fit(train_data, dev_data, submodel_method='partial_fit')

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        if args.predict_proba:
            log.warning('This model does not output probability estimates')

        original_df = deepcopy(ds.data_frame)
        length = sum(ds.encoded_ds_lenghts) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=[f'prediction_{i}' for i in range(self.horizon)])

        if self.use_stl and self.ts_analysis.get('stl_transforms', False):
            ds.data_frame = _stl_transform(ydf, ds, self.target, self.tss, self.ts_analysis)

        for timestep in range(self.horizon):
            ydf[f'prediction_{timestep}'] = self.models[timestep](ds, args)['prediction'].values

        if self.use_stl and self.ts_analysis.get('stl_transforms', False):
            ydf = _stl_inverse_transform(ydf, ds, self.tss, self.ts_analysis)

        if self.models[0].positive_domain:
            ydf = ydf.clip(0)

        ydf['prediction'] = ydf.values.tolist()
        ds.data_frame = original_df
        return ydf[['prediction']]



from copy import deepcopy
from typing import Dict, Tuple, List, Union

import optuna
import numpy as np
import pandas as pd
from sktime.transformations.series.detrend import Detrender
from sktime.forecasting.trend import PolynomialTrendForecaster
from sktime.transformations.series.detrend import ConditionalDeseasonalizer

from lightwood.api.types import TimeseriesSettings
from type_infer.dtype import dtype
from lightwood.helpers.ts import get_ts_groups, get_delta, get_group_matches, Differencer
from lightwood.helpers.log import log
from lightwood.encoder.time_series.helpers.common import generate_target_group_normalizers


def timeseries_analyzer(data: Dict[str, pd.DataFrame], dtype_dict: Dict[str, str],
                        timeseries_settings: TimeseriesSettings, target: str) -> Dict:
    """
    This module analyzes (pre-processed) time series data and stores a few useful insights used in the rest of Lightwood's pipeline.
    
    :param data: dictionary with the dataset split into train, val, test subsets. 
    :param dtype_dict: dictionary with inferred types for every column.
    :param timeseries_settings: A `TimeseriesSettings` object. For more details, check `lightwood.types.TimeseriesSettings`.
    :param target: name of the target column.
    
    The following things are extracted from each time series inside the dataset:
      - group_combinations: all observed combinations of values for the set of `group_by` columns. The length of this list determines how many time series are in the data.
      - deltas: inferred sampling interval 
      - ts_naive_residuals: Residuals obtained from the data by a naive forecaster that repeats the last-seen value. 
      - ts_naive_mae: Mean residual value obtained from the data by a naive forecaster that repeats the last-seen value.
      - target_normalizers: objects that may normalize the data within any given time series for effective learning. See `lightwood.encoder.time_series.helpers.common` for available choices.
    
    :return: Dictionary with the aforementioned insights and the `TimeseriesSettings` object for future references.
    """  # noqa
    tss = timeseries_settings
    groups = get_ts_groups(data['train'], tss)
    deltas, periods, freqs = get_delta(data['train'], dtype_dict, groups, target, tss)

    normalizers = generate_target_group_normalizers(data['train'], target, dtype_dict, groups, tss)

    if dtype_dict[target] in (dtype.integer, dtype.float, dtype.num_tsarray):
        naive_forecast_residuals, scale_factor = get_grouped_naive_residuals(data['dev'], target, tss, groups)
        differencers = get_differencers(data['train'], target, groups, tss.group_by)
        stl_transforms = get_stls(data['train'], data['dev'], target, periods, groups, tss)
    else:
        naive_forecast_residuals, scale_factor = {}, {}
        differencers = {}
        stl_transforms = {}

    return {'target_normalizers': normalizers,
            'deltas': deltas,
            'tss': tss,
            'group_combinations': groups,
            'ts_naive_residuals': naive_forecast_residuals,
            'ts_naive_mae': scale_factor,
            'periods': periods,
            'sample_freqs': freqs,
            'stl_transforms': stl_transforms,
            'differencers': differencers
            }


def get_naive_residuals(target_data: pd.DataFrame, m: int = 1) -> Tuple[List, float]:
    """
    Computes forecasting residuals for the naive method (forecasts for time `t` is the value observed at `t-1`).
    Useful for computing MASE forecasting error.

    As per arxiv.org/abs/2203.10716, we resort to a constant forecast based on the last-seen measurement across the entire horizon.
    By following the original measure, the naive forecaster would have the advantage of knowing the actual values whereas the predictor would not.

    Note: method assumes predictions are all for the same group combination. For a dataframe that contains multiple
     series, use `get_grouped_naive_resiudals`.

    :param target_data: observed time series targets
    :param m: season length. the naive forecasts will be the m-th previously seen value for each series

    :return: (list of naive residuals, average residual value)
    """  # noqa
    # @TODO: support categorical series as well
    residuals = np.abs(target_data.values[1:] - target_data.values[0]).flatten()
    scale_factor = np.average(residuals)
    return residuals.tolist(), scale_factor


def get_grouped_naive_residuals(
        info: pd.DataFrame,
        target: str,
        tss: TimeseriesSettings,
        group_combinations: List) -> Tuple[Dict, Dict]:
    """
    Wraps `get_naive_residuals` for a dataframe with multiple co-existing time series.
    """  # noqa
    group_residuals = {}
    group_scale_factors = {}
    for group in group_combinations:
        idxs, subset = get_group_matches(info, group, tss.group_by)
        if subset.shape[0] > 1:
            residuals, scale_factor = get_naive_residuals(subset[target])  # @TODO: pass m once we handle seasonality
            group_residuals[group] = residuals
            group_scale_factors[group] = scale_factor
    return group_residuals, group_scale_factors


def get_differencers(data: pd.DataFrame, target: str, groups: List, group_cols: List):
    differencers = {}
    for group in groups:
        idxs, subset = get_group_matches(data, group, group_cols)
        differencer = Differencer()
        differencer.fit(subset[target].values)
        differencers[group] = differencer
    return differencers


def get_stls(train_df: pd.DataFrame,
             dev_df: pd.DataFrame,
             target: str,
             sps: Dict,
             groups: list,
             tss: TimeseriesSettings
             ) -> Dict[str, object]:
    stls = {'__default': None}
    for group in groups:
        if group != '__default':
            _, tr_subset = get_group_matches(train_df, group, tss.group_by)
            _, dev_subset = get_group_matches(dev_df, group, tss.group_by)
            if tr_subset.shape[0] > 0 and dev_subset.shape[0] > 0 and sps.get(group, False):
                group_freq = tr_subset['__mdb_inferred_freq'].iloc[0]
                tr_subset = deepcopy(tr_subset)[target]
                dev_subset = deepcopy(dev_subset)[target]
                tr_subset.index = pd.date_range(start=tr_subset.iloc[0], freq=group_freq,
                                                periods=len(tr_subset)).to_period()
                dev_subset.index = pd.date_range(start=dev_subset.iloc[0], freq=group_freq,
                                                 periods=len(dev_subset)).to_period()
                stl = _pick_ST(tr_subset, dev_subset, sps[group])
                log.info(f'Best STL decomposition params for group {group} are: {stl["best_params"]}')
                stls[group] = stl
    return stls


def _pick_ST(tr_subset: pd.Series, dev_subset: pd.Series, sp: list):
    """
    Perform hyperparam search with optuna to find best combination of ST transforms for a time series.

    :param tr_subset: training series used for fitting blocks. Index should be datetime, and values are the actual time series.
    :param dev_subset: dev series used for computing loss. Index should be datetime, and values are the actual time series.
    :param sp: list of candidate seasonal periods
    :return: best deseasonalizer and detrender combination based on dev_loss
    """  # noqa

    def _ST_objective(trial: optuna.Trial):
        trend_degree = trial.suggest_categorical("trend_degree", [1])
        ds_sp = trial.suggest_categorical("ds_sp", sp)  # seasonality period to use in deseasonalizer
        if min(min(tr_subset), min(dev_subset)) <= 0:
            decomp_type = trial.suggest_categorical("decomp_type", ['additive'])
        else:
            decomp_type = trial.suggest_categorical("decomp_type", ['additive', 'multiplicative'])

        detrender = Detrender(forecaster=PolynomialTrendForecaster(degree=trend_degree))
        deseasonalizer = ConditionalDeseasonalizer(sp=ds_sp, model=decomp_type)
        transformer = STLTransformer(detrender=detrender, deseasonalizer=deseasonalizer, type=decomp_type)
        transformer.fit(tr_subset)
        residuals = transformer.transform(dev_subset)

        trial.set_user_attr("transformer", transformer)
        return np.power(residuals, 2).sum()

    space = {"trend_degree": [1, 2], "ds_sp": sp, "decomp_type": ['additive', 'multiplicative']}
    study = optuna.create_study(sampler=optuna.samplers.GridSampler(space))
    study.optimize(_ST_objective, n_trials=8)

    return {
        "transformer": study.best_trial.user_attrs['transformer'],
        "best_params": study.best_params
    }


class STLTransformer:
    def __init__(self, detrender: Detrender, deseasonalizer: ConditionalDeseasonalizer, type: str = 'additive'):
        """
        Class that handles STL transformation and inverse, given specific detrender and deseasonalizer instances.
        :param detrender: Already initialized. 
        :param deseasonalizer: Already initialized. 
        :param type: Either 'additive' or 'multiplicative'.
        """  # noqa
        self._type = type
        self.detrender = detrender
        self.deseasonalizer = deseasonalizer
        self.op = {
            'additive': lambda x, y: x - y,
            'multiplicative': lambda x, y: x / y
        }
        self.iop = {
            'additive': lambda x, y: x + y,
            'multiplicative': lambda x, y: x * y
        }

    def fit(self, x: Union[pd.DataFrame, pd.Series]):
        self.deseasonalizer.fit(x)
        self.detrender.fit(self.op[self._type](x, self.deseasonalizer.transform(x)))

    def transform(self, x: Union[pd.DataFrame, pd.Series]):
        return self.detrender.transform(self.deseasonalizer.transform(x))

    def inverse_transform(self, x: Union[pd.DataFrame, pd.Series]):
        return self.deseasonalizer.inverse_transform(self.detrender.inverse_transform(x))
